#ifndef SIGMAPOINT_H_
#define SIGMAPOINT_H_
#include "LinearFilters.h"
#include <math.h>
#include <unistd.h>

template <unsigned short n>
struct SIGMAPOINT_SETTINGS {
  unsigned short p; //The number of sigma points
  NT w0; //The weight on the zero vector
  NT *weights; // The weight vector
  vec<n> *sigs; // The actual sigma points
};

template <unsigned short n>
void getSigmaPoints(const SIGMAPOINT_SETTINGS<n> &sig_set, NT alpha, const GAUSSIAN<n> &rv, vec<n> *result, vec<n> *sigmas) {
  unsigned short k1;
  //
  k1 = sig_set.p;
  for (unsigned short i=0;i<k1;i++) {
    result[i] = rv.mean + rv.cov*sig_set.sigs[i]*alpha;
    if (sigmas != NULL){
      sigmas[i] = rv.cov*sig_set.sigs[i]*alpha;
    }
  }
};

template <unsigned short n, unsigned short p>
struct COMPOSITE_VECTOR {
  vec<n> v1;
  vec<p> v2;
};

template <unsigned short n, unsigned short p>
void stitch_and_root(GAUSSIAN<n+p> &comp_root, const GAUSSIAN<n> &rvar1, const GAUSSIAN<p> &rvar2){
  mat<n,n> u1,vt1;
  vec<n> sig1;
  mat<n,n> root1;
  mat<p,p> u2,vt2;
  vec<p> sig2;
  mat<p,p> root2;

  comp_root.mean = join(rvar1.mean,rvar2.mean);

  svd(rvar1.cov,u1,sig1,vt1);
  for (unsigned short i=0;i<n;i++){
    if(sig1[i]>1E-20 || sig1[i]<-1E-20){
      sig1[i] = sqrt(sig1[i]);
    } else {
      sig1[i] = 0;
    }
  }
  root1 = u1*diag(sig1)*vt1;

  svd(rvar2.cov,u2,sig2,vt2);
  for (unsigned short i=0;i<p;i++){
    if(sig2[i]>1E-20 || sig2[i]<1E-20){
      sig2[i] = sqrt(sig2[i]);
    } else {
      sig2[i] = 0;
    }
  }
  root2 = u2*diag(sig2)*vt2;

  comp_root.cov = direct_sum(root1,root2);

};


template <unsigned short n, unsigned short p, typename U>
ADVANCEMENT<n,p> sigmaAdvance(const SIGMAPOINT_SETTINGS<n+p> &sp_set, const NT alpha, ADVPROTO<n,p,U> &linp) {
  ADVANCEMENT<n,p> rval;
  COMPOSITE_VECTOR<n,p> *comp_points_s;
  COMPOSITE_VECTOR<n,p> *comp_points_c;
  GAUSSIAN<n+p> comp_rvector;
  vec<n> y0,y0c;
  vec<n> *ys,*yc;
  vec<n> ym;
  mat<n,n> pyx;
  mat<n,p> pyw;
  //Generate the Sigma Points for the composite random variable
  stitch_and_root(comp_rvector, linp.state, linp.noise);

  comp_points_s = (COMPOSITE_VECTOR<n,p>*)malloc(sp_set.p*sizeof(COMPOSITE_VECTOR<n,p>));
  comp_points_c = (COMPOSITE_VECTOR<n,p>*)malloc(sp_set.p*sizeof(COMPOSITE_VECTOR<n,p>));

  getSigmaPoints(sp_set,alpha,comp_rvector,(vec<n+p>*)comp_points_s,(vec<n+p>*)comp_points_c);

  y0 = linp.fun(NULL,NULL,linp.state.mean,linp.noise.mean,linp.forcing);
  //Pass through the Sigma Points
  ys = (vec<n>*)malloc(sp_set.p*sizeof(vec<n>));
  for (int i=0;i<sp_set.p;i++){
    ys[i] = linp.fun(NULL,NULL,comp_points_s[i].v1,comp_points_s[i].v2,linp.forcing);
  }
  //Find the mean of the new random variable
  ym = y0*sp_set.w0;
  for (int i=0;i<sp_set.p;i++){
    ym = ym + ys[i]*sp_set.weights[i];
  }
  //Recenter the sigma points about the mean
  y0c = y0-ym;
  yc = (vec<n>*)malloc(sp_set.p*sizeof(vec<n>));
  for (int i=0;i<sp_set.p;i++){
    yc[i] = ys[i] - ym;
  }
  //Find the covariance matrix Pyx and use it to calculate the matrix H
  for (int i=0;i<sp_set.p;i++){
    pyx = pyx + (outer(yc[i],comp_points_c[i].v1))*(sp_set.weights[i]);
  }
  rval.F = pyx * pinv(linp.state.cov) / (alpha*alpha);
  //Find the covariance matrix Pyw and use it to calculate the matrix V
  for (int i=0;i<sp_set.p;i++) {
    pyw = pyw + outer(yc[i],comp_points_c[i].v2)*(sp_set.weights[i]);
  }
  rval.L = pyw * pinv(linp.noise.cov) / (alpha*alpha);

  rval.x = ym - rval.F*linp.state.mean - rval.L*linp.noise.mean;

  //This is Sigma Point so we have a nonlinear covariance contribution
  rval.havenonlin = true;
  rval.nonlin = outer(y0c,y0c)*(sp_set.w0);
  for (int i=0;i<sp_set.p;i++){
    rval.nonlin = rval.nonlin + outer(yc[i],yc[i])*(sp_set.weights[i]);
  }
  rval.nonlin = rval.nonlin / (alpha*alpha) - (rval.F*linp.state.cov*trans(rval.F) + rval.L*linp.noise.cov*trans(rval.L));
  free(comp_points_s);
  free(comp_points_c);
  free(ys);
  free(yc);
  return rval;
};

template <unsigned short n, unsigned short p, unsigned short t, typename U>
MEASUREMENT<n,p,t> sigmaMeasurement(const SIGMAPOINT_SETTINGS<n+t> &sp_set, NT alpha, MEASPROTO<n,p,t,U> &linp) {
  MEASUREMENT<n,p,t> rval;
  COMPOSITE_VECTOR<n,t> *comp_points_s;
  COMPOSITE_VECTOR<n,t> *comp_points_c;
  GAUSSIAN<n+t> comp_rvector;
  vec<p> y0,y0c;
  vec<p> *ys,*yc;
  vec<p> ym;
  mat<p,n> pyx;
  mat<p,t> pyw;
  //Generate the Sigma Points for the composite random variable
  stitch_and_root(comp_rvector, linp.state, linp.noise);

  /*
    printf("   MAT\n");
    showMat(comp_rvector.cov);
    printf("   MAT1\n");
    showMat(linp.state.cov);
    printf("   MAT2\n");
    showMat(linp.noise.cov);
  */

  comp_points_s = (COMPOSITE_VECTOR<n,t>*)malloc(sp_set.p*sizeof(COMPOSITE_VECTOR<n,t>));
  comp_points_c = (COMPOSITE_VECTOR<n,t>*)malloc(sp_set.p*sizeof(COMPOSITE_VECTOR<n,t>));

  getSigmaPoints(sp_set,alpha,comp_rvector,(vec<n+t>*)comp_points_s,(vec<n+t>*)comp_points_c);
  /*
    for (int i=0;i<sp_set.p;i++){
    printf("  Vector\n");
    showVec(comp_points_c[i].v1);
    showVec(comp_points_c[i].v2);
    }
  */
  y0 = linp.fun(NULL,NULL,linp.state.mean,linp.noise.mean,linp.forcing);
  //Pass through the Sigma Points
  ys = (vec<p>*)malloc(sp_set.p*sizeof(vec<p>));
  for (int i=0;i<sp_set.p;i++){
    ys[i] = linp.fun(NULL,NULL,comp_points_s[i].v1,comp_points_s[i].v2,linp.forcing);
  }
  //Find the mean of the new random variable
  ym = y0*sp_set.w0;
  for (int i=0;i<sp_set.p;i++){
    ym = ym + ys[i]*sp_set.weights[i];
  }
  //Recenter the sigma points about the mean
  y0c = y0-ym;
  yc = (vec<p>*)malloc(sp_set.p*sizeof(vec<p>));
  for (int i=0;i<sp_set.p;i++){
    yc[i] = ys[i] - ym;
  }
  //Find the covariance matrix Pyx and use it to calculate the matrix H
  for (int i=0;i<sp_set.p;i++){
    pyx = pyx + (outer(yc[i],comp_points_c[i].v1))*(sp_set.weights[i]);
  }
  rval.H = pyx * pinv(linp.state.cov) / (alpha*alpha);
  //Find the covariance matrix Pyw and use it to calculate the matrix V
  for (int i=0;i<sp_set.p;i++) {
    pyw = pyw + outer(yc[i],comp_points_c[i].v2)*(sp_set.weights[i]);
  }
  rval.V = pyw * pinv(linp.noise.cov) / (alpha*alpha);

  rval.y = ym - rval.H*linp.state.mean - rval.V*linp.noise.mean;

  //This is Sigma Point so we have a nonlinear covariance contribution
  rval.havenonlin = true;
  rval.nonlin = outer(y0c,y0c)*(sp_set.w0);
  for (int i=0;i<sp_set.p;i++){
    rval.nonlin = rval.nonlin + outer(yc[i],yc[i])*(sp_set.weights[i]);
  }

  rval.nonlin = rval.nonlin / (alpha*alpha) - (rval.H*linp.state.cov*trans(rval.H) + rval.V*linp.noise.cov*trans(rval.V));
  free(comp_points_s);
  free(comp_points_c);
  free(ys);
  free(yc);
  return rval;
};

template<unsigned short d>
void Generate_Simplex_Set(SIGMAPOINT_SETTINGS<d> &rval) {
  mat<d+1,d+1> prototype;
  //Find the prototype first
  //The weight vector
  for (int k=0; k<d+1;k++) {
    prototype[0][k] = 1/sqrt(d+1);
  }
  //Find the other values
  for (int k=d; k>0;k--) {
    //Set the first element
    NT accounted_for = prototype[0][k]*prototype[0][k];
    for (int j=(k+1);j<d+1;j++){
      accounted_for = accounted_for + prototype[j][k]*prototype[j][k];
    }
    prototype[k][k] = sqrt(1-accounted_for);
    accounted_for = 1-accounted_for;
    for (int j=0;j<k;j++){
      prototype[k][j]=sqrt( (1-accounted_for)/k );
    }
  }
  showMat(prototype);
  rval.p = d+1;
  //Turn the prototype into a matrix
  rval.w0 = 0;
  rval.weights = (NT*)malloc((d+1)*sizeof(NT));
  for (int k=0;k<d+1;k++){
    rval.weights[k]=prototype[0][k]*prototype[0][k];
  }
  rval.sigs = (vec<d>*)malloc((d+1)*sizeof(vec<d>));
  for (int k=1;k<d+1;k++){
    rval.sigs[k-1] = vec<d>();
    for (int j=0;j<d+1;j++){
      rval.sigs[k-1][j] = prototype[k][j]/prototype[0][j];
    }
  }
}

template<unsigned short d>
void Generate_Order3_Set(SIGMAPOINT_SETTINGS<d> &rval){
  rval.p = 2*d;
  rval.w0 = 1-(NT)d;
  rval.weights = (NT*)malloc(2*d*sizeof(NT));
  for (int k=0;k<2*d;k++){
    rval.weights[k] = 0.5;
  }
  rval.sigs = (vec<d>*)malloc(2*d*sizeof(vec<d>));
  for (int k=0;k<d;k++){
    rval.sigs[2*k+0] = vec<d>();
    rval.sigs[2*k+0][k] = 1.0;
    rval.sigs[2*k+1] = vec<d>();
    rval.sigs[2*k+1][k] = -1.0;
  }
}

template<unsigned short d>
void Generate_Order5_Set(SIGMAPOINT_SETTINGS<d> &rval) {
	//Set the weights
	rval.p = 2*d;
	rval.w0 = (2.0*((NT)d)/3.0)-((NT)d)+1.0;
	rval.weights = (NT*)malloc(2*d*sizeof(NT));
	for (int k=0; k<2*d; k++)	{
		rval.weights[k] = 1.0/6.0;
	}
	//Set the sigma points
	rval.sigs = (vec<d>*)malloc(2*d*sizeof(vec<d>));
	for (int k=0; k<d; k++){
    rval.sigs[2*k+0] = vec<d>();
		rval.sigs[2*k+0][k]=sqrt(3.0);
    rval.sigs[2*k+1] = vec<d>();
		rval.sigs[2*k+1][k]=-sqrt(3.0);
	}
}

template<unsigned short d>
void Generate_Order7_Set(SIGMAPOINT_SETTINGS<d> &rval) {
	rval.p = 4*d;
	rval.w0 = 1-(NT)d;
	rval.weights = (NT*)malloc(4*d*sizeof(NT));
	for (int k=0; k<d; k++){
		rval.weights[4*k+0]=0.0458758547681;
		rval.weights[4*k+1]=0.454124145232;
		rval.weights[4*k+2]=0.454124145232;
		rval.weights[4*k+3]=0.0458758547681;
	}
	rval.sigs = (vec<d>*)malloc(4*d*sizeof(vec<d>));
	for (int k=0;k<d;k++){
    rval.sigs[4*k+0] = vec<d>();
		rval.sigs[4*k+0][k]=-2.33441421834;
    rval.sigs[4*k+1] = vec<d>();
		rval.sigs[4*k+1][k]=-0.741963784303;
    rval.sigs[4*k+2] = vec<d>();
		rval.sigs[4*k+2][k]=0.741963784303;
    rval.sigs[4*k+3] = vec<d>();
		rval.sigs[4*k+3][k]=2.3344142183;
	}
}

template<unsigned short d>
void Generate_Order9_Set(SIGMAPOINT_SETTINGS<d> &rval) {
  rval.p = 4*d;
  rval.w0 = 0.533333333333333*((NT)d)+1-(NT)d;
  rval.weights = (NT*)malloc(4*d*sizeof(NT));
  for (int k=0; k<d; k++){
    rval.weights[4*k+0]=0.0112574113277207;
    rval.weights[4*k+1]=0.222075922005613;
    rval.weights[4*k+2]=0.222075922005613;
    rval.weights[4*k+3]=0.0112574113277207;
  }
  rval.sigs = (vec<d>*)malloc(4*d*sizeof(vec<d>));
  for (int k=0;k<d;k++){
    rval.sigs[4*k+0] = vec<d>();
    rval.sigs[4*k+0][k]=2.85697001387281;
    rval.sigs[4*k+1] = vec<d>();
    rval.sigs[4*k+1][k]=1.35562617997427;
    rval.sigs[4*k+2] = vec<d>();
    rval.sigs[4*k+2][k]=-1.35562617997427;
    rval.sigs[4*k+3] = vec<d>();
    rval.sigs[4*k+3][k]=-2.85697001387281;
  }
}

template <int n>
struct factorial {
	enum { value = n * factorial<n - 1>::value };
};

template <>
struct factorial<0> {
	enum { value = 1 };
};

template <int n, int m>
struct comb {
  enum { value = factorial<n>::value/(factorial<m>::value * factorial<n-m>::value)};
};

template<unsigned short d>
void Generate_Order5_Full_Set(SIGMAPOINT_SETTINGS<d> &rval) {
  unsigned short npairs = (comb<d,2>::value);
  unsigned short pairs[npairs][2];
  unsigned short t_ind;
  NT ntd;
  ntd = (NT)d;
  rval.p = 4*npairs + 2*d + 2*d;
  rval.w0 = 1-npairs+2.0/3.0*ntd-ntd+(ntd-1)*ntd;
  rval.weights = (NT*)malloc(rval.p*sizeof(NT));
  for (int k=0;k<npairs;k++){
    rval.weights[4*k+0] =0.25;
    rval.weights[4*k+1] =0.25;
    rval.weights[4*k+2] =0.25;
    rval.weights[4*k+3] =0.25;
  }
  for (int k=0;k<d;k++){
    rval.weights[npairs*4+2*k+0] =0.16666667;
    rval.weights[npairs*4+2*k+1] =0.16666667;
  }
  for (int k=0;k<d;k++){
    rval.weights[npairs*4+2*d+2*k+0] = -0.5*(d-1);
    rval.weights[npairs*4+2*d+2*k+1] = -0.5*(d-1);
  }
  rval.sigs = (vec<d>*)malloc(rval.p*sizeof(vec<d>));
  t_ind = 0;
  for (int j=0;j<d;j++){
    for (int k=(j+1);k<d;k++){
      pairs[t_ind][0] = j;
      pairs[t_ind][1] = k;
      t_ind++;
    }
  }
  for (int k=0;k<npairs;k++){
    rval.sigs[4*k+0] = vec<d>();
    rval.sigs[4*k+0][pairs[k][0]] =1;
    rval.sigs[4*k+0][pairs[k][1]] =1;
    rval.sigs[4*k+1] = vec<d>();
    rval.sigs[4*k+1][pairs[k][0]] =1;
    rval.sigs[4*k+1][pairs[k][1]] =-1;
    rval.sigs[4*k+2] = vec<d>();
    rval.sigs[4*k+2][pairs[k][0]] =-1;
    rval.sigs[4*k+2][pairs[k][1]] =1;
    rval.sigs[4*k+3] = vec<d>();
    rval.sigs[4*k+3][pairs[k][0]] =-1;
    rval.sigs[4*k+3][pairs[k][1]] =-1;
  }
  for (int k=0;k<d;k++){
    rval.sigs[npairs*4+2*k+0] = vec<d>();
    rval.sigs[npairs*4+2*k+0][k] =1.73205080757;
    rval.sigs[npairs*4+2*k+1] = vec<d>();
    rval.sigs[npairs*4+2*k+1][k] =-1.73205080757;
  }
  for (int k=0;k<d;k++){
    rval.sigs[npairs*4+2*d+2*k+0] = vec<d>();
    rval.sigs[npairs*4+2*d+2*k+0][k] =1;
    rval.sigs[npairs*4+2*d+2*k+1] = vec<d>();
    rval.sigs[npairs*4+2*d+2*k+1][k] =-1;
  }
  /*if(npairs<40){
    printf("%d points \n",rval.p);
    printf("npairs = %d\n",npairs);
    printf("w0 %f\n",rval.w0);
    for (int k=0;k<rval.p;k++){
    printf("%f ",rval.weights[k]);
    showVec(rval.sigs[k]);
    }
    }*/
}

//Still need to look at the pairwise set

template<unsigned short d>
void delete_set(SIGMAPOINT_SETTINGS<d> &rval){
  free(rval.weights);
  free(rval.sigs);
}
#endif /* SIGMAPOINT_H_ */
