#ifndef LINALG_H_
#define LINALG_H_

#include "linalg.h"

#define USE_DOUBLE //Need to do it this way for lapack

#define PINV_MIN 1E-20

#ifdef USE_DOUBLE
typedef double NT;
#else
typedef float NT;
#endif

#include <stdio.h>
extern "C" {
	#include <lapacke.h>
}
#include <math.h>
template <unsigned short n> struct vec {
	NT val[n];

	// Construction from 0
	inline 	vec(const NT s = NT(0) ) {
		for (unsigned short i=0;i<n;i++){
			val[i] = s;
		}
	}

	// Construction from float array
	inline 	vec(const NT inval[n]){
		for (unsigned short i=0;i<n;i++){
			val[i]=inval[i];
		}
	}

	// Construction from vector
	inline 	vec(const vec<n> &v){
		for (unsigned short i=0;i<n;i++){
			val[i]=v.val[i];
		}
	}

	// Unary Minus
	inline 	vec<n> operator - () const {
		vec<n> rval;
		for (unsigned short i=0;i<n;i++){
			rval[i]=-val[i];
		}
		return rval;
	}

	// Binary Addition
	inline 	vec<n> operator + (const vec<n> &v) const {
		vec<n> rval;
		for (unsigned short i=0;i<n;i++){
			rval[i] = val[i] + v[i];
		}
		return rval;
	}

	// Binary Subtraction
	inline 	vec<n> operator - (const vec<n> &v) const {
		vec<n> rval;
		for (unsigned short i=0;i<n;i++){
			rval[i] = val[i]-v[i];
		}
		return rval;
	}

	// Binary Inner Product
	inline 	NT operator * (const vec<n> &v) const {
		NT rval = 0;
		for (unsigned short i=0;i<n;i++){
			rval = rval + val[i]*v[i];
		}
		return rval;
	}

	// Binary Inner Product with a Vector prototype
	/*NT operator * (const NT v[n]) const {
		NT rval = 0;
		for (unsigned short i=0;i<n;i++){
			rval = rval + val[i]*v[i];
		}
		return rval;
	}*/

	// Binary Scale Down Operation
	inline 	vec<n> operator / (const NT s) const {
		vec<n> rval;
		for (unsigned short i=0;i<n;i++) {
			rval[i] = val[i]/s;
		}
		return rval;
	}

	// Binary Scale Up Operation
	inline 	vec<n> operator * (const NT s) const {
		vec<n> rval;
		for (unsigned short i=0;i<n;i++) {
			rval[i] = val[i]*s;
		}
		return rval;
	}

	inline 	operator const NT* () const {
		return val;
	}

	inline 	operator NT* () {
		return val;
	}
};
template < unsigned short n>
inline vec<n> dtimes(const vec<n> &a, const vec<n> &b) {
	vec<n> rval;
	for (unsigned short i=0;i<n;i++){
		rval[i] = a[i]*b[i];
	}
	return rval;
}


template < unsigned short m,unsigned short n> struct mat {
	NT val[m*n];

	// Construction from 0
	inline 	mat(NT s = NT(0.0) ) {
		for (unsigned short i=0;i<m*n;i++){
			val[i] = 0;
		}
		if (s != 0.0) {
			if (m>n) {
				for (unsigned short i=0;i<n;i++){
					val[n*i+i] = s;
				}
			} else {
				for (unsigned short i=0;i<m;i++){
					val[n*i+i] = s;
				}
			}
		}
	}

	// Construct from 1D float array
	inline 	mat(const NT inval[m*n]) {
		for (unsigned short i=0;i<m*n;i++){
			val[i]=inval[i];
		}
	}

	// Construct from 2D float array
	inline 	mat(const NT inval[m][n]) {
		for (unsigned short m_ind=0;m_ind<m;m_ind++){
			for (unsigned short n_ind=0;n_ind<n;n_ind++){
				val[n*m_ind+n_ind]=inval[m_ind][n_ind];
			}
		}
	}

	// Construct from 1D array of vectors
	inline 	mat(const vec<n> v[m]){
		for (unsigned short m_ind=0;m_ind<m;m_ind++){
			for (unsigned short n_ind=0;n_ind<n;n_ind++){
				val[n*m_ind+n_ind]=v[m_ind].val[n_ind];
			}
		}
	}

	// Construct from another matrix
	inline 	mat(const mat<m,n> &v){
		for (unsigned short i=0;i<m*n;i++){
			val[i]=v.val[i];
		}
	}

	inline 	NT* operator [] (unsigned short m_ind) {
		return &val[n*m_ind];
	}
	inline 	NT const * operator [] (unsigned short m_ind) const {
		return &val[n*m_ind];
	}

	// Transpose
	inline 	mat<n,m> operator ! (){
		mat<n,m> rval;
		for (unsigned short m_ind=0;m_ind<m;m_ind++){
			for (unsigned short n_ind=0;n_ind<n;n_ind++){
				rval.val[n_ind*m+m_ind]=val[m_ind*n+n_ind];
			}
		}
		return rval;
	}

	// Matrix Vector Multiply
	inline 	vec<m> operator * (const vec<n> &v) const {
		vec<m> rval;
		vec<n> tval;
		for (unsigned short m_ind=0;m_ind<m;m_ind++){
			rval[m_ind] = v * &val[m_ind*n];
		}
		return rval;
	}

	// Unary Negative
	inline 	mat<m,n> operator - () const {
		mat<m,n> rval;
		for (unsigned short i=0;i<m*n;i++){
			rval.val[i] = -val[i];
		}
		return rval;
	}

	// Matrix Matrix Addition
	inline 	mat<m,n> operator + (const mat<m,n> &m2) const {
		mat<m,n> rval;
		for (unsigned short i=0;i<m*n;i++){
			rval.val[i] = val[i]+m2.val[i];
		}
		return rval;
	}

	// Matrix Matrix Subtraction
	inline 	mat<m,n> operator - (const mat<m,n> &m2) const {
		mat<m,n> rval;
		for (unsigned short i=0;i<m*n;i++){
			rval.val[i] = val[i] - m2.val[i];
		}
		return rval;
	}

	// Matrix Scaling
	inline 	mat<m,n> operator * (const NT s) const {
		mat<m,n> rval;
		for (unsigned short i=0;i<m*n;i++){
			rval.val[i] = s*val[i];
		}
		return rval;
	}

	// Matrix Scaling
	inline mat<m,n> operator / (const NT s) const {
		mat<m,n> rval;
		for (unsigned short i=0;i<m*n;i++) {
			rval.val[i] = val[i]/s;
		}
		return rval;
	}
};

template < unsigned short n>
inline mat<n,n> diag(const vec<n> &v) {
	mat<n,n> rval;
	for (unsigned short i=0;i<n;i++) {
		rval.val[i*n+i] = v[i];
	}
	return rval;
}

// Simple Transpose
template < unsigned short m, unsigned short n>
inline mat<m,n> trans(const mat<n,m> &m1) {
	mat<m,n> rval;
	for (unsigned short m_ind=0;m_ind<m;m_ind++){
		for (unsigned short n_ind=0;n_ind<n;n_ind++) {
			rval.val[m_ind*n+n_ind] = m1[n_ind][m_ind];
		}
	}
	return rval;
}

// Matrix matrix Multiply
template < unsigned short m,unsigned short n,unsigned short p>
inline mat<m,p> operator * (const mat<m,n> &m1,const mat<n,p> &m2){
	mat<m,p> rval;
	for (unsigned short m_ind=0;m_ind<m;m_ind++){
		for (unsigned short p_ind=0;p_ind<p;p_ind++){
			for (unsigned short n_ind=0;n_ind<n;n_ind++){
				rval.val[m_ind*p+p_ind]=rval[m_ind][p_ind]+m1[m_ind][n_ind]*m2[n_ind][p_ind];
			}
		}
	}
	return rval;
};

// How to display a matrix
template < unsigned short m,unsigned short n>
inline void showMat(const mat<m,n> &v){
	for (unsigned short m_ind=0;m_ind < m;m_ind++){
		printf("[");
		for(unsigned short n_ind=0;n_ind<n;n_ind++){
			printf("%+e,",v.val[m_ind*n+n_ind]);
		}
		printf("]\n");
	}
};

// How to display a vector
template < unsigned short n>
inline void showVec(const vec<n> &v){
	printf("(");
	for (unsigned short n_ind=0;n_ind<n;n_ind++){
		printf("%e,",v[n_ind]);
	}
	printf(")\n");
};

// How to join vectors
template < unsigned short n, unsigned short m>
inline vec<n+m> join(const vec<n> &v1,const vec<m> &v2){
	vec<n+m> rval;
	unsigned short n_ind;
	for (n_ind = 0;n_ind<n;n_ind++){
		rval[n_ind]=v1[n_ind];
	}
	for (unsigned short m_ind = 0;m_ind<m;m_ind++){
		rval[n_ind]=v2[m_ind];
		n_ind++;
	}
	return rval;
};

// How to direct sum
template < unsigned short n1, unsigned short n2, unsigned short p1, unsigned short p2>
inline mat<n1+n2,p1+p2> direct_sum(const mat<n1,p1> &m1, const mat<n2,p2> &m2){
  mat<n1+n2,p1+p2> rval;
  unsigned short n1_ind,n2_ind;
  for(n1_ind=0;n1_ind<n1;n1_ind++){
    for(n2_ind=0;n2_ind<p1;n2_ind++){
      rval[n1_ind][n2_ind] = m1[n1_ind][n2_ind];
    }
  }
  for(n1_ind=0;n1_ind<n2;n1_ind++){
    for(n2_ind=0;n2_ind<p2;n2_ind++){
      rval[n1_ind+n1][n2_ind+p1] = m2[n1_ind][n2_ind];
    }
  }
  return rval;
};

template <unsigned short n1, unsigned short p, unsigned short n2>
inline mat<n1+n2,p> stack(const mat<n1,p> &m1, const mat<n2,p> &m2){
  mat<n1+n2,p> rval;
  unsigned short n1_ind,n2_ind;
  for(n1_ind=0;n1_ind<n1;n1_ind++){
    for(n2_ind=0;n2_ind<p;n2_ind++){
      rval[n1_ind][n2_ind] = m1[n1_ind][n2_ind];
    }
  }
  for(n1_ind=0;n1_ind<n2;n1_ind++){
    for(n2_ind=0;n2_ind<p;n2_ind++){
      rval[n1_ind+n1][n2_ind]=m2[n1_ind][n2_ind];
    }
  }
}


template <unsigned short n,unsigned short p>
inline mat<n,p> outer(const vec<n> &v1, const vec<p> &v2) {
	mat<n,p> rval;
	mat<1,n> mv1 = v1.val;
	mat<1,p> mv2 = v2.val;
	rval = trans(mv1)*mv2;
	return rval;
};

#ifdef USE_DOUBLE
template <unsigned short n>
inline mat<n,n> inv(const mat<n,n> &a){
	mat<n,n> rval = a;
	int ipiv[n+1];
	LAPACKE_dgetrf(LAPACK_ROW_MAJOR,n,n,rval.val,n,ipiv);
	LAPACKE_dgetri(LAPACK_ROW_MAJOR,n,rval.val,n,ipiv);
	return rval;
};
#else
template <unsigned short n>
mat<n,n> inv(const mat<n,n> &a){
	mat<n,n> rval = a;
	int ipiv[n+1];
	LAPACKE_sgetrf(LAPACK_ROW_MAJOR,n,n,rval.val,n,ipiv);
	LAPACKE_sgetri(LAPACK_ROW_MAJOR,n,rval.val,n,ipiv);
	return rval;
}
#endif

#ifdef USE_DOUBLE
template <unsigned short n>
inline mat<n,n> msqrt(const mat<n,n> &a){
	mat<n,n> rval = a;
	LAPACKE_dpotrf(LAPACK_ROW_MAJOR, 'U', n, rval.val, n);
	return rval;
};
#else
#endif

#ifdef USE_DOUBLE

// NOTE m is assumed to be greater than or equal to n
template <unsigned short m, unsigned short n>
inline mat<n,m> pinv(const mat<m,n> &A) {
	vec<n> s;
	mat<m,n> u;
	mat<n,n> vt;
	mat<m,n> tval = A;
	mat<n,m> rval;
	double suprb[n+1];

	// if (n>m) {
	// 	printf("Bad matrix for pinvert doing the transpose myself, %d,%d\n",m,n);
	// 	rval = trans(pinv(trans(tval)));
	// 	return rval;
	// }

	LAPACKE_dgesvd(LAPACK_ROW_MAJOR,'S','S',m,n,tval.val,n,s.val,u.val,n,vt.val,n,suprb);

	for (unsigned short i=0;i<n;i++) {
		if (s[i]>PINV_MIN || s[i]<-PINV_MIN) {
			s[i] = 1/s[i];
		} else {
			s[i] = 0;
		}
	}

	rval = trans(vt)*diag(s)*trans(u);
	return rval;
};
#else
#endif

#ifdef USE_DOUBLE
template <unsigned short n>
inline void svd(const mat<n,n> &A, mat<n,n> &u, vec<n> &s, mat<n,n> &vt) {
	mat<n,n> tval = A;
	double suprb[n+1];
	LAPACKE_dgesvd(LAPACK_ROW_MAJOR,'A','A',n,n,tval.val,n,s.val,u.val,n,vt.val,n,suprb);
}
#else
#endif

#ifdef USE_DOUBLE
template <unsigned short n>
vec<n> lusolve(const mat<n,n> &A,const vec<n> &v) {
	vec<n> sol;
	mat<n,n> mval;
	int ipiv[n+1];
	mval = A;
	sol = v;

	LAPACKE_dgesv(LAPACK_ROW_MAJOR,n,1,mval.val,n,ipiv,sol.val,n);

	return sol;
}
#else
template <unsigned short n>
vec<n> lusolve(const mat<n,n> &A,const vec<n> &v) {
	vec<n> sol;
	mat<n,n> mval;
	int ipiv[n+1];
	mval = A;
	sol = v;

	LAPACKE_sgesv(LAPACK_ROW_MAJOR,n,1,mval.val,n,ipiv,sol.val,n);

	return sol;
}
#endif

// How to take the determinant of a matrix
template < unsigned short n>
NT det(const mat<n,n> &v){
	NT rval;
	unsigned short permute[n];
	mat<n,n> L;
	mat<n,n> U;
	rval = 0;
	switch(lupDecompU(L,U,permute,v)){
	case 0:
		rval = 0;
		break;
	case 2: // Matrix singular
		rval = 0;
		break;
	case 1:
		rval = 1;
		for (unsigned short i=0;i<n;i++){
			if(permute[i] != i){
				rval = -rval;
			}
		}
		for (unsigned short i=0;i<n;i++)
			rval = rval*U[i][i];
		break;
	}
	return rval;
};

// Perform a LU decomposition, with U having non 1 diagonal
// From Numerical Analysis by Burden, Faires - Thomson
template < unsigned short n>
unsigned short lupDecompU(mat<n,n> &l,mat<n,n> &u,unsigned short *permute, const mat<n,n> &v){
	mat<n,n> a;
	unsigned short trow;
	NT scratch;
	unsigned short i,j,k;
	unsigned short s;

	a = v;
	trow = 0;
	for (i=0;i<n;i++){
		permute[i] = i;
	}
	while (a[trow][0] == 0){ // Figure out a permutation that allows the decomp
		trow++;
		if (trow==n){
			return 0; // No way to decompose it
		}
	}

	for (i=0;i<n;i++){
		scratch = a[0][i];
		a[0][i]=a[trow][i];
		a[trow][i]=scratch;
	}
	permute[trow]=0;
	permute[0]=trow;

	u[0][0]=a[0][0];
	l[0][0]=1;

	for(j=1;j<n;j++){
		u[0][j] = a[0][j];
		l[j][0] = a[j][0] / u[0][0];
	}

	for(i=1;i<(n-1);i++){
		trow = i;
		scratch = 0;
		for (k=0;k<i;k++){
			scratch = scratch + l[i][k]*u[k][i];
		}
		while ( (a[trow][i]-scratch) == 0 ) {
			trow++;
			if (trow == n){
				return 0;
			}
		}
		u[i][i] = a[trow][i] - scratch;
		l[i][i] = 1;
		for(j=0;j<n;j++){
			scratch = a[i][j];
			a[i][j]=a[trow][j];
			a[trow][j]=scratch;
		}

		s = permute[i];
		permute[i]=permute[trow];
		permute[trow]=s;

		for(j=(i+1);j<n;j++){
			scratch = 0;
				for(k=0;k<i;k++){
				scratch = scratch + l[i][k]*u[k][j];
			}
			u[i][j] = a[i][j] - scratch;
			scratch = 0;
			for(k=0;k<i;k++){
				scratch = scratch + l[j][k]*u[k][i];
			}
			l[j][i] = (a[j][i] - scratch) / u[i][i];
		}
	}
	scratch = 0;
	for (k=0;k<(n-1);k++){
		scratch = scratch + l[n-1][k]*u[k][n-1];
	}
	u[n-1][n-1] = a[n-1][n-1] - scratch;
	l[n-1][n-1] = 1;
	if (u[n-1][n-1] == 0){
		return 2; // Factorization is correct but singular
	}

	return 1;
}


#endif /* LINALG_H_ */
